#define ARGS actuals result resume

open Array;;
open Utils;;
open Values;;

let current_input = ref (IPort stdin);;
let current_output = ref (OPort stdout);;

(* A builtin function expects the arguments:
   actuals : vector of function arguments
   result  : vector of result values
   resume  : continuation *)

let aux_add a b = match a, b with
  | n, Int m -> n + m
  | _ -> internal_error "invalid addition on non-int types"
;; 

let aux_mul a b = match a, b with
  | n, Int m -> n * m
  | _ -> internal_error "invalid multiplication on non-int types"
;; 

(* -------------------------- *)
(* The actual implementations *)
(* -------------------------- *)

let integer_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | Int _ -> true
       | _ -> false);
  resume ()
;;

let plus ARGS = 
  check_types (check_int "plus") actuals;
  result.(0) <- mkInt (fold_left aux_add 0 actuals);
  resume ()
;;

let times ARGS = 
  check_types (check_int "times") actuals;
  result.(0) <- mkInt (fold_left aux_mul 1 actuals);
  resume ()
;;

(* Macro for stating binary primitives over integers.
   If not stated explicitly they will be heap allocated. *)
#define BIN_OP_INT(name, s, op, mk)         \
let name ARGS =                             \
  match actuals.(0), actuals.(1) with       \
    | Int n, Int m ->                       \
	result.(0) <- mk (op n m);          \
	resume ()                           \
    | Int _, t -> type_error s (Int 0) t    \
    | t    , _ -> type_error s (Int 0) t    \
;;
BIN_OP_INT(minus,     "minus",     (-),   mkInt)
BIN_OP_INT(quotient,  "quotient",  (/),   mkInt)
BIN_OP_INT(remainder, "remainder", (mod), mkInt)
BIN_OP_INT(lt,        "<",         (<),   mkBool)
BIN_OP_INT(leq,       "<=",        (<=),  mkBool)
BIN_OP_INT(eq,        "=",         (=),   mkBool)
BIN_OP_INT(geq,       ">=",        (>=),  mkBool)
BIN_OP_INT(gt,        ">",         (>),   mkBool)

let boolean_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | True
       | False -> true
       | _ -> false);
  resume ()
;;

let symbol_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | Sym _ -> true
       | _ -> false);
  resume ()
;;

let char_p ARGS =
  result.(0) <- mkBool 
    (match actuals.(0) with
    | Char _ -> true 
    | _ -> false);
  resume ()

;;

let char_to_integer ARGS =
  match actuals.(0) with
    | Char c -> 
	result.(0) <- mkInt (Char.code c);
	resume ()
    | t -> type_error "char->integer" (Char 'a') t 
;;

let integer_to_char ARGS =
  match actuals.(0) with
    | Int i -> 
	result.(0) <- mkChar (Char.chr i);
	resume ()
    | t -> type_error "integer->char" (Int 0) t 
;;

let string ARGS = 
  let xs = String.create (Array.length actuals) in
    Array.iteri
      (fun i a -> match a with
	 | Char c -> xs.[i] <- c
	 | t -> type_error "string" (Char 'a') t)
      actuals;
    result.(0) <- Str xs;
    resume ()
;;

let make_string ARGS =
  match actuals.(0), actuals.(1) with
    | Int s, Char c ->
	result.(0) <- mkStr (String.make s c);
	resume ()
    | Int _, t -> type_error "make-string" (Char ' ') t
    | t,     _ -> type_error "make-string" (Int 0) t
;;

let string_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | Str _ -> true
       | _ -> false);
  resume ()
;;

let string_length ARGS =
  match actuals.(0) with
    | Str s ->
	result.(0) <- mkInt (String.length s);
	resume ()
    | t -> type_error "string-length" (Str "") t
;;

let string_append ARGS =
  match actuals.(0), actuals.(1) with
    | Str s1, Str s2 ->
	let l1 = String.length s1 in
	let l2 = String.length s2 in
	let ds = String.create (l1 + l2) in
	  String.blit s1 0 ds 0  l1;
	  String.blit s2 0 ds l1 l2;
	  result.(0) <- mkStr ds;
	  resume ()
    | Str _, t -> type_error "string-append" (Str "") t
    | t,     _ -> type_error "string-append" (Str "") t
;;

let string_eq_p ARGS =
  match actuals.(0), actuals.(1) with
    | Str s1, Str s2 ->
	result.(0) <- mkBool (String.compare s1 s2 = 0);
	resume ()
    | Str _, t -> type_error "string=?" (Str "") t
    | t,     _ -> type_error "string=?" (Str "") t
;;

let string_ref ARGS =
  match actuals.(0), actuals.(1) with
    | Str s, Int i ->
	result.(0) <- mkChar s.[i]; (* ocaml does the bounds checking *)
	resume ()
    | Str _, t -> type_error "string-ref" (Int 0) t
    | t,     _ -> type_error "string-ref" (Str "") t
;;

let string_to_symbol ARGS =
  match actuals.(0) with
    | Str s -> result.(0) <- Sym s; resume ()
    | t -> type_error "string->symbol" (Str "") t 
;;

let symbol_to_string ARGS =
  match actuals.(0) with
    | Sym s -> result.(0) <- Str s; resume ()
    | t -> type_error "symbol->string" (Sym "") t 
;;

let pair_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | Pair _ -> true
       | _ -> false);
  resume ()
;;

let cons ARGS =
  result.(0) <- mkPair actuals.(0) actuals.(1);
  resume ()
;;

let car ARGS =
  match actuals.(0) with
    | Pair (h,_) -> 
	result.(0) <- !h;
	resume ()
    | t -> type_error "car" (mkPair Nil Nil) t
;;

let cdr ARGS =
  match actuals.(0) with
    | Pair (_,t) -> 
	result.(0) <- !t;
	resume ()
    | t -> type_error "cdr" (mkPair Nil Nil) t
;;

let set_car_d ARGS =
  match actuals.(0) with
    | Pair (h,_) ->
	h := actuals.(1);
	result.(0) <- mkVoid;
	resume ()
    | t -> type_error "set-car!" (mkPair Nil Nil) t
;;

let set_cdr_d ARGS =
  match actuals.(0) with
    | Pair (_,t) ->
	t := actuals.(1);
	result.(0) <- mkVoid;
	resume ()
    | t -> type_error "set-cdr!" (mkPair Nil Nil) t
;;

let null_p ARGS = 
  result.(0) <- mkBool 
    (match actuals.(0) with
       | Nil -> true
       | _ -> false);
  resume ()
;;

let vector ARGS =
  result.(0) <- mkVec (Array.copy actuals);
  resume ()
;;

let make_vector ARGS =
  match actuals.(0) with
    | Int s ->
	result.(0) <- mkVec (Array.make s actuals.(1));
	resume ()
    | t -> type_error "make-vector" (Vec [||]) t
;;

let vector_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | Vec _ -> true
       | _ -> false);
  resume ()
;;

let vector_length ARGS =
  match actuals.(0) with
    | Vec v ->
	result.(0) <- mkInt (Array.length v);
	resume ()
    | t -> type_error "vector-length" (Vec [||]) t
;;

let vector_ref ARGS =
  match actuals.(0), actuals.(1) with
    | Vec v, Int i ->
	result.(0) <- v.(i);	   (* ocaml will do bounds checking *)
	resume ()
    | Vec _, t -> type_error "vector-ref" (Int 0) t
    | t,     _ -> type_error "vector-ref" (Vec [||]) t
;;

let vector_set_d ARGS =
  match actuals.(0), actuals.(1) with
    | Vec v, Int i ->
	v.(i) <- actuals.(2);	   (* ocaml will do bounds checking *)
	result.(0) <- Void;
	resume ()
    | Vec _, t -> type_error "vector-set!" (Int 0) t
    | t,     _ -> type_error "vector-set!" (Vec [||]) t
;;

let procedure_p ARGS =
  result.(0) <- mkBool
    (match actuals.(0) with
       | Clo _
       | Prim _
       | Apply
       | CallCC
       | Cont _ -> true
       | _ -> false);
  resume ()
;;

(* Special function.
   apply is recognized as the 'Apply' constructor of the values type.
   The implementation is left to the interpreter.
let apply ARGS =
  unsupported_error "apply"
;;
*)

let eqv_p ARGS =
  result.(0) <- mkBool (val_eq actuals.(0) actuals.(1));
  resume ()
;;

(* Special function.
   call/cc is recognized as the 'CallCC' constructor of the values type.
   The implementation is left to the interpreter.
let call_cc ARGS =
  unsupported_error "call_cc"
;;
*)

(* calling exit halts the interpreter since the resume continuation is
   not envoked *)
let exit ARGS =
  actuals.(0)
;;

let open_input_file ARGS =
  match actuals.(0) with 
    | Str s ->
	let ch = mkIPort (open_in s) in
	  result.(0) <- ch;
	  !current_input <- ch;
	  resume ()
    | t -> type_error "open-input-file" (Str "") t 
;;

let input_port_p ARGS = 
  result.(0) <- mkBool 
    (match actuals.(0) with
       | IPort _ -> true 
       | _ -> false);
  resume ()
;;

let close_input_port ARGS = 
  match actuals.(0) with
    | IPort ch -> 
	close_in ch;
	result.(0) <- mkVoid;
	!current_input <- IPort stdin;
	resume ()
    |t -> type_error "close-input-port" !current_input t 
;;

let current_input_port ARGS =
  result.(0) <- !current_input;
  resume ()
;;

let read_char ARGS = 
  match actuals.(0) with
    | IPort ch -> 
	(try 
	   result.(0) <- mkChar (input_char ch)	   
	 with End_of_file ->
	   result.(0) <- mkEOF);
	resume ()
    | t -> type_error "read-char" !current_input t 
;;

let peek_char ARGS =
  match actuals.(0) with
    | IPort ch -> 
	let pos = pos_in ch in
	  (try 
	     result.(0) <- mkChar (input_char ch);
	     seek_in ch pos;
	   with End_of_file ->
	     seek_in ch pos;
	     result.(0) <- mkEOF);
	  resume ()
    | t -> type_error "peek-char" !current_input t 
;;

let eof_object_p ARGS =
  result.(0) <- mkBool 
    (match actuals.(0) with
       | EOF -> true 
       | _ -> false);
  resume ()
;;

let open_output_file ARGS =
  match actuals.(0) with 
    | Str s ->
	let ch = mkOPort (open_out s) in
	  result.(0) <- ch;
	  !current_output <- ch;
	  resume ()
    | t -> type_error "open-output-file" (Str "") t 
;;

let output_port_p ARGS =
  result.(0) <- mkBool 
    (match actuals.(0) with
       | OPort _ -> true 
       | _ -> false);
  resume ()
;;

let close_output_port ARGS =
  match actuals.(0) with
    | OPort ch -> 
	close_out ch;
	result.(0) <- mkVoid;
	!current_input <- OPort stdout;
	resume ()
    |t -> type_error "close-output-port" !current_output t 
;;

let current_output_port ARGS =
  result.(0) <- !current_output;
  resume ()
;;

let write_char ARGS =
  match actuals.(0), actuals.(1) with
    | Char c, OPort ch -> 
	output_char ch c;
	result.(0) <- mkVoid;
	resume ()
    | s,t -> type_error "write-char" !current_output t 
;;

let lib : value array =
  [| mkPrim "integer?" 		   1  integer_p
  ; mkPrim "plus"     		 (-1) plus
  ; mkPrim "minus"    		   2  minus
  ; mkPrim "times"    		 (-1) times
  ; mkPrim "quotient" 		   2  quotient
  ; mkPrim "remainder" 		   2  remainder
  ; mkPrim "<" 		   	   2  lt
  ; mkPrim "<=" 		   2  leq
  ; mkPrim "=" 		   	   2  eq
  ; mkPrim ">=" 		   2  geq
  ; mkPrim ">" 		   	   2  gt
  ; mkPrim "boolean?" 		   1  boolean_p
  ; mkPrim "symbol?" 		   1  symbol_p
  ; mkPrim "char?" 		   1  char_p
  ; mkPrim "char->integer" 	   1  char_to_integer
  ; mkPrim "integer->char" 	   1  integer_to_char
  ; mkPrim "string" 		 (-1) string
  ; mkPrim "make-string" 	   2  make_string
  ; mkPrim "string?" 		   1  string_p
  ; mkPrim "string-length" 	   1  string_length
  ; mkPrim "string-append" 	   2  string_append
  ; mkPrim "string-eq?" 	   2  string_eq_p
  ; mkPrim "string-ref" 	   2  string_ref
  ; mkPrim "string->symbol" 	   1  string_to_symbol
  ; mkPrim "symbol->string" 	   1  symbol_to_string
  ; mkPrim "pair?" 		   1  pair_p
  ; mkPrim "cons" 		   2  cons
  ; mkPrim "car" 		   1  car
  ; mkPrim "cdr" 		   1  cdr
  ; mkPrim "set-car!" 		   2  set_car_d
  ; mkPrim "set-cdr!" 		   2  set_cdr_d
  ; mkPrim "null?" 		   1  null_p
  ; mkPrim "vector" 		 (-1) vector
  ; mkPrim "make-vector" 	   2  make_vector
  ; mkPrim "vector?" 		   1  vector_p
  ; mkPrim "vector-length" 	   1  vector_length
  ; mkPrim "vector-ref" 	   2  vector_ref
  ; mkPrim "vector-set!" 	   3  vector_set_d
  ; mkPrim "procedure?" 	   1  procedure_p
  ; Apply  (* mkPrim "apply"	   2  apply *)
  ; mkPrim "eqv?" 		   2  eqv_p
  ; CallCC (* mkPrim "call/cc"	   1  call_cc *)
  ; mkPrim "exit" 		   1  exit
  ; mkPrim "open-input-file" 	   1  open_input_file
  ; mkPrim "input-port?" 	   1  input_port_p
  ; mkPrim "close-input-port" 	   1  close_input_port
  ; mkPrim "current-input-port"    0  current_input_port
  ; mkPrim "read-char" 		   1  read_char
  ; mkPrim "peek-char" 		   1  peek_char
  ; mkPrim "eof-object?" 	   1  eof_object_p
  ; mkPrim "open-output-file" 	   1  open_output_file
  ; mkPrim "output-port?" 	   1  output_port_p
  ; mkPrim "close-output-port" 	   1  close_output_port
  ; mkPrim "current-output-port"   0  current_output_port
  ; mkPrim "write-char" 	   2  write_char
  |]
;;
