#include "defines.h"

open Utils;;
open Emitter;;

exception CodeGenError;;
exception CodeGenUnsupported;;

type mark = Opcode of int | Jump of int;;

let current : Values.value option ref = ref None;;

(* Special registers for DS components.
 * These registers are initialized in jit.c and must be invariant in
 * the compiled code. That means that before calling any external code
 * the values must be saved and restored after the call.
 *
 * Registers that are free for use are: eax, ebx and ecx
 *)
let reg_vec = esi;;
let reg_env = edi;;
let reg_work = ebx;;

(* Static C functions and global variables (defined in jit.c) *)
let statics = ref 0;;
let static_res    () = !statics + 0*4;;
let static_tmp    () = !statics + 1*4;;
let static_glo    () = !statics + 2*4;;
let static_lib    () = !statics + 3*4;;

let atom          () = !statics + 4*4;;
let modify        () = !statics + 5*4;;
let mkValue       () = !statics + 6*4;;
let print         () = !statics + 7*4;;
let slowMkTuple   () = !statics + 8*4;;
let fastMkTuple   () = !statics + 9*4;;

(* Move a static value in memory to a register. *)
let genStatic f r e =
  e $ mov (Int (f ())) r
    $ mov (operand_mem r) r
;;

(* Generate a call to a static C-function.
 * Notice: Uses eax as the working register as it will be overwritten
 * with the result of the call anyway.
 *)
let genCall f e =
  e $ mov (Int (f ())) eax 
    $ call (operand_mem eax)
;;

(* Debugging helper to print a register or value *)
let debug t e =
  e $ push_pop [t; reg_env; reg_vec; eax] (genCall print)
;;

(* Helper to put in a tagged ocaml integer *)
let oint i = Int (i lsl 1 lor 1);;

(* Generate a return call
 * This escapes all the way to the ocaml-interpreter.
 *)
let genRet e =
  e $ mov ebp esp	      (* restore the previous stack pointer *)
    $ pop ebp		      (* restore the base pointer *)
    $ ret 0		      (* return *)
;;

(* Generate a mkValue call.
 * This uses a static C-function that performs a callback to the
 * mkValue function in OCaml.
 * eax : resulting value
 *
 * C call convention: 
 * - args are pushed right to left
 * - eax contains return value
 * - caller clears argument stack
 * (- callee handles the stack operations)
 *)
let genMkValue t d e = match t with
  | TYPE_NIL  -> e $ mov (Int VAL_NIL) eax
  | TYPE_BOOL -> e $ mov (Int (if d=0 then VAL_FALSE else VAL_TRUE)) eax
  | TYPE_VOID -> e $ mov (Int VAL_VOID) eax
  | _ ->
      e $ push_pop [Int t; Int d; reg_env; reg_vec] (genCall mkValue)
;;

(* Check for small allocation sizes *)
let small s = s <= 256;;

(* Allocate a tuple that may contain garbage.
 * eax : resulting tuple
 *)
let genMkTupleUnsafe size e = match size with
  | 0 -> e $ genStatic atom eax (* we may not call alloc for size zero *)
  | _ ->
      e $ push_pop [Int size; reg_env; reg_vec]
	  (genCall (if small size then fastMkTuple else slowMkTuple))
;;

(* Allocate a tuple initialized with the DS value VAL_VOID
 * eax : resulting tuple
 *)
let genMkTuple size e =
  let e = e $ genMkTupleUnsafe size in
    if small size then
      e $ mov (Int VAL_VOID) reg_work
	$ repeati size (fun i -> mov reg_work (operand_disp eax (i*4)))
    else e
;;

(* Modify example *)
let genModify d i v e =
  (* this must be nested as 'd' could be reg_env or reg_vec *)
  e $ push_pop [reg_env; reg_vec]
    (fun e ->
       e $ add (Int (i*4)) d
	 $ push_pop [d; v] (genCall modify))
;;

(* TODO:
(* Generate a fast call to a builtin function *)
(* match !current with
   | Some (Values.Prim (n, a, f)) -> e $ genBuiltinCall (n, a, f)
   | _ -> ... *)
let genBuiltinCall (n, a, f) e = match f with
(*| f when f == Builtins.plus -> ... do somthing ... *)
  | _ -> e $ push_pop [Int (address f); reg_env; reg_vec] (genCall callBuiltin)
;;
*)

(* Generate code to write a value to some DS vector.
 * eax : the value to put somewhere
 * reg_work : will be invalidated
 *)
let genPut s i e = match s with
  | REG_LIB ->
      e $ genStatic static_lib reg_work
	$ genModify reg_work i eax
  
  | REG_GLO ->
      e $ genStatic static_glo reg_work
	$ genModify reg_work i eax

  | REG_TMP -> 
      e $ genStatic static_tmp reg_work
	$ genModify reg_work i eax

  | REG_RES ->
      e $ genStatic static_res reg_work
	$ genModify reg_work i eax
	
  | REG_VEC ->
      e $ genModify reg_vec i eax

  | n ->
      e $ mov reg_env reg_work
	$ repeat n (mov (operand_disp reg_work 4) reg_work) (* cd...dr *)
	$ mov (operand_mem reg_work) reg_work		    (* car *)
	$ genModify reg_work i eax
;;

(* Generate code to read a value from some DS vector.
 * eax : the resulting read value
 * reg_env : must point to env
 * reg_vec : must point to vec
 * No other registers are invalidated
 *)
let genGet s i e = match s with
  | REG_LIB ->
      e $ genStatic static_lib eax
	$ mov (operand_disp eax (i*4)) eax

  | REG_GLO ->
      e $ genStatic static_glo eax
	$ mov (operand_disp eax (i*4)) eax

  | REG_TMP -> 
      e $ genStatic static_tmp eax
	$ mov (operand_disp eax (i*4)) eax

  | REG_RES ->
      e $ genStatic static_res eax
	$ mov (operand_disp eax (i*4)) eax
	
  | REG_VEC ->
      e $ mov (operand_disp reg_vec (i*4)) eax

  | n ->
      e $ mov reg_env eax
	$ repeat n (mov (operand_disp eax 4) eax) (* cd...dr *)
	$ mov (operand_mem  eax) eax		  (* car *)
	$ mov (operand_disp eax (i*4)) eax
;;

let rec genOpcodes bits e = 
(*
  let e = bitmatch bits with
    | { op : 8 } -> e $ debug (Int op)
    | { _ } -> e
  in
*)
  bitmatch bits with

    | { OP_NOP : 8
      ; bits : -1 : bitstring }
      ->
	e $ mark (Opcode 1)
	  $ nop
	  $ genOpcodes bits

    | { OP_RETURN : 8 
      ; bits : -1 : bitstring }
      ->
	e $ mark (Opcode 1)
	  (* when size <= 256 we can perform unsafe direct assignments
	     assuming that no allocations take place between
	     assignments *)
	  $ genMkTupleUnsafe 5 
	  $ mov (oint 0) (operand_mem eax)

	  (* clo, vec, env, and size not set! But since we never read
	     from them it should be ok? maybe they should be
	     initialized to something... *)

	  (* hmmm, good question. It could be a problem because the
	     garbage collector might look at the values, which are
	     random data, and see something that looks like a pointer
	     thereby failing to reclaim some storage... However, the
	     return value is in eax and it is garbage immidiatly after
	     returning to OCaml. No garbage collection happens between
	     returing and the subsequent tail-call to continue so the
	     tuple itself should be collected and possible pointers in
	     it would not matter... *)

	  $ genRet
	  $ genOpcodes bits

    | { OP_CALL : 8
      ; scope : 8
      ; index : 16 : nativeendian
      ; size  : 16 : nativeendian
      ;	bits  : -1 : bitstring }
      ->
	e $ mark (Opcode 6)
	  $ genMkTupleUnsafe 5
	  (* fst = pc *)
	  $ delay (fun late -> mov (oint (pc late)) (operand_mem eax))
	  (* snd = clo *)
	  $ mov eax reg_work	    (* temporary move tuple to work *)
	  $ genGet scope index
	  $ mov eax (operand_disp reg_work (1*4))
	  $ mov reg_work eax	    (* restore the tuple to eax *)
	  (* thd = vec *)
	  $ mov reg_vec  (operand_disp eax (2*4))
	  (* four = env *)
	  $ mov reg_env  (operand_disp eax (3*4))
	  (* five = size *)
	  $ mov (oint size) (operand_disp eax (4*4))
	  (* 5-tuple is the return value *)
	  $ genRet
	  (* set the correct (resume) pc and continue *)
	  $ force
	  $ genOpcodes bits

    | { OP_TAIL_CALL : 8
      ; scope : 8 
      ; index : 16 : nativeendian
      ; bits  : -1 : bitstring }
      ->
	e $ mark (Opcode 4)
	  $ genMkTupleUnsafe 5
	  $ mov eax reg_work
	  (* fst = 1 *)
	  $ mov (oint 1) (operand_mem reg_work)
	  (* snd = clo *)
	  $ genGet scope index
	  $ mov eax (operand_disp reg_work (1*4))
	  (* thd = vec *)
	  $ mov reg_vec (operand_disp reg_work (2*4))

	  (* env and size not set! But since we never read from them it should be ok?
	     maybe they should be initialized to something... *)

	  (* set return value *)
	  $ mov reg_work eax
	  $ genRet
	  $ genOpcodes bits

    | { OP_LOAD : 8
      ; vtype : 8 
      ; data  : 32 : nativeendian
      ; scope : 8
      ; index : 16 : nativeendian
      ; bits  : -1 : bitstring }
      ->
	e $ mark (Opcode 9)
	  $ genMkValue vtype (Int32.to_int data)
	  $ genPut scope index
   	  $ genOpcodes bits

    | { OP_MOVE : 8
      ; sscope : 8
      ; sindex : 16 : nativeendian
      ; tscope : 8
      ; tindex : 16 : nativeendian
      ; bits   : -1 : bitstring } 
      -> 
	e $ mark (Opcode 7)
	  $ genGet sscope sindex
	  $ genPut tscope tindex
	  $ genOpcodes bits

    | { OP_NEW_VEC : 8
      ; size : 16 : nativeendian
      ; bits : -1 : bitstring } 
      ->
	e $ mark (Opcode 3)
	  $ genMkTuple size
	  $ mov eax reg_vec		     (* set new vec *)
	  $ genOpcodes bits

    | { OP_EXTEND : 8
      ; bits : -1 : bitstring }
      ->
	e $ mark (Opcode 1)
	  $ genMkTupleUnsafe 2		     (* create a pair *)
	  (* vec -> car *)
	  $ mov reg_vec (operand_mem eax)
	  (* env -> cdr *)
	  $ mov reg_env (operand_disp eax (1*4))
	  (* set new env *)
	  $ mov eax reg_env
	  $ genOpcodes bits

    | { OP_JUMP : 8 
      ; offset : 32 : nativeendian
      ; bits   : -1 : bitstring }
      ->
	e $ mark (Opcode 5)
	  $ jmp (Int 0)
	  $ mark (Jump (Int32.to_int offset))
	  $ genOpcodes bits

    | { OP_JUMP_IF_FALSE : 8
      ; scope  : 8
      ; index  : 16 : nativeendian
      ; offset : 32 : nativeendian, bind (Int32.to_int offset)
      ; bits   : -1 : bitstring } 
      ->
	e $ mark (Opcode 8)
	  $ genGet scope index
	  $ sub (Int 7) eax
	  $ jcc Zero (Int 0)
	  $ mark (Jump offset)
	  $ genOpcodes bits

    | { bits : -1 : bitstring, check(Bitstring.bitstring_length bits = 0) }
      -> e

    | { _ }
      -> raise CodeGenUnsupported;
;;

let genClosure body loc e =

  let rec opaddr i ms = match i, ms with  
    | _, [] -> internal_error "jumped outside of code"
    | 0, (Opcode j, pc)::ms -> pc
    | i, (Opcode j, pc)::ms when i < j -> internal_error "address has not been marked"
    | i, (Opcode j, pc)::ms -> opaddr (i-j) ms
    | i, m::ms -> opaddr i ms
  in
  
  let dojump e =
    let ms = List.rev (marks e) in
    let rec visit = function
      | [] -> ()
      | (Jump ds, pc)::js ->
	  let jmpt = opaddr (ds - loc) ms - pc in
	  let _ = e $ set_pc (pc - 4) $ emit_int jmpt in
	    visit js
      | j::js -> visit js
    in visit ms; e

  in e $ genOpcodes body
       $ dojump
;;
