#include "defines.h"

open Utils;;
open Values;;
open Bitstring;;

(* External just-in-time compilation functions (defined in jit.c). *)
type jit_result = int * value * value array * env * int;;
external jit_exec : string -> int -> value array -> env -> jit_result = "exec";;
external jit_init : value array -> value array -> value array -> value array -> int = "init";;

(* Jit compilation type for memoization *)
type jit_comp = JitUndefined | JitUnsupported | JitCompiled of string;;

(* Flags to control tracing and jit compilation (triggered from main.cpp.ml). *)
let trace_level = ref 0;;
let enable_jit = ref false;;
let jit_stats = ref false;;

(* Information collection for jit compilation. *)
let jit_executed = ref 0;;
let jit_exec_inc () =
#ifndef RELEASE
  if !jit_stats then jit_executed := !jit_executed + 1;
#endif
  ()
;;
let jit_print_stats code lam jl e =
#ifndef RELEASE
  if !jit_stats then
    let rec count u c bmin bmax lmin lmax rmin rmax = function
      | -1 -> (u, c, bmin, bmax, lmin, lmax, rmin, rmax)
      | i  -> (match jl.(i) with
		 | JitUnsupported -> count (u+1) c bmin bmax lmin lmax rmin rmax
		 | JitCompiled bs  ->
		     let b = String.length bs in
		     let l = (if i < Array.length lam - 1 then snd lam.(i+1)
			      else bitstring_length code) - snd lam.(i) in
		     let r = b / l in
		       count u (c+1)
			 (min bmin b)
			 (max bmax b)
			 (min lmin l)
			 (max lmax l)
			 (min rmin r)
			 (max rmax r)
		 | _ -> count u c bmin bmax lmin lmax rmin rmax) (i-1)
    in
    let elen = String.length !(e.Emitter.buffer) in
    let u, c, bmin, bmax, lmin, lmax, rmin, rmax =
      count 0 0 elen 0 elen 0 elen 0 (Array.length jl - 1) in
      fprintf std_formatter
	"Jit Statistics
 closures total: %d
 unsupported:    %d
 compiled:       %d
 executions:     %d
 min closure:    %d bytes
 max closure:    %d bytes
 min compiled:   %d bytes
 max compiled:   %d bytes
 emitter size    %d bytes
 min ratio:      %d
 max ratio:      %d@."
	(Array.length jl)
	u c !jit_executed
	lmin lmax
	bmin bmax elen
	rmin rmax;
#endif
  ()
;;

(* Main interpreter (invoked from main.cpp.ml). *)

let interpret (maxglo, maxtmp, maxres,
	       str, sym, lam, code, sign) =
  
  let glo = Array.make maxglo Void in
  let tmp = Array.make maxtmp Void in
  let res = Array.make maxres Nil in
  let lib = Builtins.lib in

  let code_length = bitstring_length code in
  let bits_to_pc bits =
    (code_length - bitstring_length bits) / 8
  in

  (* maps a register tag to its structure *)
  let get id env vec = match id with
    | REG_LIB -> lib
    | REG_GLO -> glo 
    | REG_RES -> res
    | REG_TMP -> tmp
    | REG_VEC -> vec
    | n       -> List.nth env n in

  (* setup dynamic compilation stuff *)
  let max_lam = let rec find h = function
    | i when i < 0 -> h
    |  i -> find (max h (snd lam.(i+1) - snd lam.(i))) (i-1)
  in if !enable_jit then find 1 (Array.length lam - 2) else 0 in
  (* scaling by 11 is chosen by emperical tests *)
  let jitbuf = Emitter.make (max_lam * 11) in
  let jitlam = Array.make (if !enable_jit
			   then Array.length lam
			   else 0) JitUndefined in
  let jitclo i env = match jitlam.(i) with
    | JitUnsupported   -> None
    | JitCompiled comp -> Some (Comp (fst lam.(i), comp, env))
    | JitUndefined ->
	(* try compiling the closure *)
	let (arity, offset) = lam.(i) in
	let offset_end =
	  if i+1 < Array.length lam
	  then snd lam.(i+1) * 8
	  else bitstring_length code
	in
	  try
	    let comp = Emitter.contents
	      (Codegen.genClosure
		 (dropbits (offset*8) (takebits (offset_end) code))
		 offset
		 (Emitter.reset jitbuf))
	    in
	      jitlam.(i) <- JitCompiled comp;
	      Some (Comp (arity, comp, env))
	  with Codegen.CodeGenUnsupported ->
	    jitlam.(i) <- JitUnsupported;
	    None
  in
  let jit_mkValue t data env vec = match t with
    | TYPE_CLO_F | TYPE_CLO_D ->
	let env = if t = TYPE_CLO_F then [vec] else env in
	  maybe (Clo (data, env)) (jitclo data env)
    | _ -> mkValue t data env vec str sym 
  in
  if !enable_jit then
    (let _ = Callback.register "identity" id in
     let _ = Callback.register "mkValue" jit_mkValue in
     let statics = jit_init res tmp glo lib in
       Codegen.statics := statics);

  (* visit is the main loop of the interpreter *)
  let rec visit bits env vec cont =
    let (op, bits) = bitmatch bits with
      | { op : 8; bits : -1 : bitstring } -> (op, bits)
      | { _ } -> runtime_error "no more code to execute"
    in
#ifndef RELEASE
      (* TRACING START *)
      if !trace_level > 0 then (
	fprintf std_formatter "%5d %s"
	  (bits_to_pc bits)
	  (op_str op);
	if !trace_level > 1 then (
	  let sepc ff () = fprintf ff ";" in
	  let seps ff () = fprintf ff " " in
	  let arrp ff a =
	    fprintf ff "[%a]" (fprintfs sepc print) (Array.to_list a) in
	  let envp ff a =
	    fprintf ff "%a" (fprintfs seps arrp) a
	  in fprintf std_formatter
	       " env%a  tmp%a  res%a  glo%a  vec%a"
	       envp env arrp tmp arrp res arrp glo arrp vec
	);
	fprintf std_formatter "@."
      );
      (* TRACING END *)
#endif
    match op with
      | OP_NOP
	-> visit bits env vec cont

      | OP_LOAD ->
	  let (vtype, data, scope, index) = bitmatch bits with
	      { vtype : 8 
	      ; data  : 32 : nativeendian
	      ; scope : 8
	      ; index : 16 : nativeendian }
	      -> (vtype, data, scope, index)
	  in (get scope env vec).(index) <-
	       (match !enable_jit, mkValue vtype (Int32.to_int data) env vec str sym with
		  | true, (Clo (i, env) as clo) -> maybe clo (jitclo i env)
		  | _, v -> v);
	    visit (dropbits (8*8) bits) env vec cont

      | OP_MOVE ->
	  let (sscope, sindex, tscope, tindex) = bitmatch bits with
	      { sscope : 8
	      ; sindex : 16 : nativeendian
	      ; tscope : 8
	      ; tindex : 16 : nativeendian }
	      -> (sscope, sindex, tscope, tindex)
	  in (get tscope env vec).(tindex) <- (get sscope env vec).(sindex);
	    visit (dropbits (6*8) bits) env vec cont

      | OP_NEW_VEC ->
	  let size = bitmatch bits with
	      { size : 16 : nativeendian } -> size
	  in visit (dropbits (2*8) bits) env (Array.make size Void) cont

      | OP_EXTEND -> 
	  visit bits (vec::env) vec cont

      | OP_JUMP -> 
	  let offset = bitmatch bits with
	      { offset : 32 : nativeendian } -> offset
	  in visit (dropbits (Int32.to_int offset * 8) code) env vec cont

      | OP_JUMP_IF_FALSE -> 
	  let (scope, index, offset) = bitmatch bits with
	      { scope  : 8
	      ; index  : 16 : nativeendian
	      ; offset : 32 : nativeendian } -> (scope, index, offset)
	  in (match (get scope env vec).(index) with
		| False
	          -> visit (dropbits (Int32.to_int offset * 8) code) env vec cont
		| _ -> visit (dropbits (7*8) bits) env vec cont)
	       
      | OP_TAIL_CALL ->
	  let (scope, index) = bitmatch bits with 
	      { scope : 8 
	      ; index : 16 : nativeendian } -> (scope, index)
	  in application (get scope env vec).(index) vec cont

      | OP_CALL -> 
	  let (scope, index, size) = bitmatch bits with
	      { scope : 8
	      ; index : 16 : nativeendian
	      ; size  : 16 : nativeendian } -> (scope, index, size)
	  in application (get scope env vec).(index) vec
	       (Frame (bits_to_pc bits + 5, env, Array.sub tmp 0 size, cont))

      | OP_RETURN ->
	  continue cont

      | _ -> runtime_error ("unknown opcode: " ^ string_of_int op)

  (* defines the application of any applicable object *)
  and application clo vec cont = match clo with
    | Prim (n, a, f) -> 
	check_arity n a vec;
	f vec res (fun () -> continue cont)
    | Clo (i, env) ->
	let (arity, offset) = lam.(i) in
	  check_arity "user closure" arity vec;
	  let arg = if arity = -1 then Array.make 1 (vec_to_list vec) else vec in
	    visit (dropbits (offset*8) code) (arg::env) (mkInitVec ()) cont
    | Comp (arity, comp, env) ->
	check_arity "user closure" arity vec;
	let arg = if arity = -1 then Array.make 1 (vec_to_list vec) else vec in
	let env = arg::env in
	  jit_exec_inc ();
	  exec comp 0 env cont
    | Apply  ->
	check_arity "apply" 2 vec;
	check_procedure "apply" vec.(0);
	(* check_proper_list vec.(1); *)
	(* we could optimize this to avoid the packing/unpacking in
	   the case that vec.(0) is a Clo with arity = -1 *)
	application vec.(0) (list_to_vec vec.(1)) cont
    | CallCC ->
	check_arity "call/cc" 1 vec;
	check_procedure "call/cc" vec.(0);
	application vec.(0) (Array.make 1 (mkCont cont)) cont
    | Cont cf ->
	check_arity "continuation" 1 vec;
	res.(0) <- vec.(0);
	continue cf
    | _ -> runtime_error ("unapplicable value: " ^ type_str clo)

  (* defines how to continue evaluation
     (eg, after a return or applying a continuation) *)
  and continue = function
    | Init -> jit_print_stats code lam jitlam jitbuf; res.(0)
    | Frame (pc, env, stmp, cont)
      -> for i = 0 to Array.length stmp - 1 do
	   tmp.(i) <- stmp.(i)
         done;
	 visit (dropbits (pc*8) code) env (mkInitVec ()) cont
    | CFrame (comp, pc, env, stmp, cont)
      -> for i = 0 to Array.length stmp - 1 do
	   tmp.(i) <- stmp.(i)
         done;
	 exec comp pc env cont

  (* defines the transfer to and return from machine code *)
  and exec comp pc env cont =
    match jit_exec comp pc [||] env with
      | ( 0,   _,   _,   _,    _) -> continue cont
      | ( 1, clo, vec,   _,    _) -> application clo vec cont
      | (pc, clo, vec, env, size) -> application clo vec
	  (CFrame (comp, pc, env, Array.sub tmp 0 size, cont))
      | _ -> internal_error "unexpected return value from compiled code."

  in visit code [] (mkInitVec ()) Init
;;
