#include "defines.h"

open Utils;;

type index = int
type name  = string
type arity = int
type value =
  (* external value types (see defines.h) *)
  | Void
  | Nil
  | True
  | False
  | Int  of int
  | Char of char
  | Str  of string
  | Sym  of string
  | Clo  of index * env (* both flat and deep closures *)
  (* internal value types *)
  | Pair of value ref * value ref
  | Vec  of value array
  | Prim of name * arity * builtin
  | Apply
  | CallCC
  | Cont of cf
  | IPort of in_channel
  | OPort of out_channel  
  | EOF
  | Comp of int * string * env
and cf =
  | Init
  | Frame of int * env * value array * cf
  | CFrame of string * int * env * value array * cf
and env     = value array list
and resume  = unit -> value
and builtin = value array -> value array -> resume -> value
;;

let mkNil = Nil;;
let mkPair a b = Pair (ref a, ref b);;
let mkBool b = if b then True else False;;
let mkInt n = Int n;;
let mkChar c = Char c;;
let mkStr s = Str s;;
let mkSym s = Sym s;;
let mkClo i e = Clo (i,e);;
let mkVoid = Void;;
let mkVec v = Vec v;;
let mkPrim s a f = Prim (s,a,f);;
let mkCont c = Cont c;;
let mkIPort ch = IPort ch;;
let mkOPort ch = OPort ch;;
let mkEOF = EOF;;

let mkValue t data env vec str sym = match t with
  | TYPE_NIL   -> Nil
  | TYPE_BOOL  -> if data != 0 then True else False
  | TYPE_INT   -> Int data
  | TYPE_CHAR  -> Char (Char.chr data)
  | TYPE_STR   -> Str str.(data)
  | TYPE_SYM   -> Sym sym.(data)
  | TYPE_CLO_F -> Clo (data, [vec])
  | TYPE_CLO_D -> Clo (data, env)
  | TYPE_VOID  -> Void
  (* You can not make a primitive function *)
  | _ -> internal_error ("invalid type constructor: " ^ string_of_int t)
;;

let mkInitVec () = Array.make 0 Void;;

let rec val_eq a b = match a, b with
  | Int   a, Int   b when a==b -> true
  | Char  a, Char  b when a==b -> true
(*| Str   a, Str   b when a==b -> true *)
  | Sym   a, Sym   b when a=b  -> true
  | IPort a, IPort b when a==b -> true
  | OPort a, OPort b when a==b -> true
(*| Vec [||], Vec [||]         -> false *)
(*| Vec   a, Vec   b when a==b -> true *)
  | a, b -> a == b
;;

let print =
  let rec p ff = function
    | Nil -> fprintf ff "()"
    | True -> fprintf ff "#t"
    | False -> fprintf ff "#f"
    | Int n -> fprintf ff "%d" n
    | Char c -> fprintf ff "#\\%a" chr c
    | Str s -> fprintf ff "\"%s\"" s
    | Sym s -> fprintf ff "%s" s
    | Clo _ | Comp _ -> fprintf ff "clo"
    | Void -> fprintf ff "\"#<void object>\""
    | Pair _ as c -> fprintf ff "(%a)" lst c
    | Vec v -> vec ff v
    | Prim (n,_,_) -> fprintf ff "%s" n
    | Apply -> fprintf ff "apply"
    | CallCC -> fprintf ff "call/cc"
    | Cont _ -> fprintf ff "continuation"
    | IPort _ -> fprintf ff "iport"
    | OPort _ -> fprintf ff "oport"
    | EOF -> fprintf ff "eof"
  and lst ff = function
    | Pair (h, t) when !t = Nil -> fprintf ff "%a" p !h
    | Pair (h1, t1) ->
	(match !t1 with
	   | Pair _ -> fprintf ff "%a %a"   p !h1 lst !t1
	   | _      -> fprintf ff "%a . %a" p !h1 p   !t1)
    | _ -> internal_error "unexpected type in Values.print lst"
 (* Function for DAIMI-Scheme's very idiosyncratic printing of vectors.
  * #len(a b c) where if len is larger then the number of printed
  * elements it means that all remaining elements have the value 'c'.
  *)
  and vec ff v =
    let rec last_uniq i =
      if i<1 || not (val_eq v.(i) v.(i-1))
      then i
      else last_uniq (i-1)
    in let len = (Array.length v) in
      fprintf ff "#%d(%a)" len (ssep_space p)
	(Array.to_list (Array.sub v 0 (last_uniq (len - 1) + 1)))
  and chr ff = function
   (* special names (R6RS). what about #\esc and #\delete? *)
    | '\000' -> fprintf ff "nul"
    | '\007' -> fprintf ff "alarm"
    | '\008' -> fprintf ff "backspace"
    | '\009' -> fprintf ff "tab"
    | '\010' -> fprintf ff "newline"
    | '\011' -> fprintf ff "vtab"
    | '\012' -> fprintf ff "page"
    | '\013' -> fprintf ff "return"
    | '\027' -> fprintf ff "esc"
    | '\032' -> fprintf ff "space"
    | '\127' -> fprintf ff "delete"
   (* ascii chars (directly printable) *)
    | c when (Char.code c > 33 && Char.code c < 127)
	-> fprintf ff "%c" c
   (* the rest are printed in hex *)
    | c -> fprintf ff "x%X" (Char.code c)
  in p
;;

let type_str = function
  | Nil    -> "nil"
  | True
  | False  -> "bool"
  | Int  _ -> "int"
  | Char _ -> "char"
  | Str  _ -> "string"
  | Sym  _ -> "symbol"
  | Void   -> "void"
  | Pair _ -> "pair"
  | Vec _  -> "vector"
  | Cont _ -> "continutation" 
  | IPort _-> "iport"
  | OPort _-> "oport"
  | EOF    -> "eof"
  | Clo _ | Comp _ | Prim _ | Apply | CallCC -> "procedure"
;;

let vec_to_list xs =
  let l = Array.length xs in
  let rec visit i =
    if i = l
    then Nil
    else mkPair xs.(i) (visit (i+1))
  in visit 0
;;

let list_to_vec lst =
  let rec visit i = function
    | Nil -> Array.make i Nil
    | Pair (h, t) ->
	let xs = visit (i+1) !t in
	  xs.(i) <- !h;
	  xs
    | _ -> internal_error "list_to_vec should only be applied to proper lists"
  in visit 0 lst
;;

let type_error s e f =
  raise (TYPE_ERROR (error_prefix
		     ^ "expected " ^ type_str e
		     ^ ", found " ^ type_str f
		     ^ " in function " ^ s))
;;

let check_int s = function
  | Int n -> ()
  | t -> type_error s (Int 0) t
;;

let check_procedure s = function
  | Clo _ | Comp _ | Prim _ | Apply | CallCC | Cont _ -> ()
  | t -> type_error s Apply t
;;

let check_types ck xs =
  Array.iter ck xs
;;

let check_arity name arity args =
  if arity = -1 then ()
  else if arity != Array.length args
  then runtime_error (
    "arity mismatch for " ^ name
    ^ " expected " ^ string_of_int arity ^ " agruments, received "
    ^ string_of_int (Array.length args)
  )
;;

let op_str = function
  | OP_NOP           -> "nop"
  | OP_LOAD          -> "loa"
  | OP_MOVE          -> "mov"
  | OP_NEW_VEC       -> "new"
  | OP_EXTEND        -> "ext"
  | OP_JUMP          -> "jmp"
  | OP_JUMP_IF_FALSE -> "cjm"
  | OP_TAIL_CALL     -> "tcl"
  | OP_CALL          -> "cal"
  | OP_RETURN        -> "ret"
  | n -> runtime_error ("invalid opcode " ^ string_of_int n)
;;

