open Format;;
open Utils;;
open Bitstring;;

module VMLParser =
struct

  let parse_eos bits =
    bitmatch bits with
      | { 0x80 : 8; bits : -1 : bitstring } -> bits
      | { _ } -> parse_error "expected EOS"

  let parse_header bits =
    (* We first check that the magic word is ordered according to the
       native endianness of the machine.
       If all is well we return the limit values and the rest of the
       bits. *) 
    bitmatch bits with
      | { (0xDA15|0xCE03) : 16 : nativeendian
	; (0xDA15|0xCE03) : 16 : nativeendian
	; glo  : 16 : nativeendian
	; tmp  : 16 : nativeendian
	; res  : 16 : nativeendian
	; bits : -1 : bitstring
	} -> (glo, tmp, res, bits)
      | { _ } ->
	  parse_error "invalid magic header or endian mis-match"

  let parse_string_entry bits =
    bitmatch bits with
      | { len : 32 : nativeendian
	; str : (Int32.to_int len * 8) : string
	; bits : -1 : bitstring
	} -> (str, bits)
      | { _ } -> parse_error "invalid string entry"

  let parse_strings bits =
    bitmatch bits with
      | { count : 32 : nativeendian
	; bits  : -1 : bitstring
	} ->
	  let count = Int32.to_int count in
	  let arr   = Array.make count "" in
	  let rbits = ref bits in
	    for i = 0 to count - 1 do
	      let (str, bits) = parse_string_entry !rbits in
		arr.(i) <- str;
		rbits := bits
	    done;
	    (arr, parse_eos !rbits)

  let parse_lambda_entry bits =
    bitmatch bits with
      | { arity : 8  : nativeendian
	; code  : 32 : nativeendian
	; bits  : -1 : bitstring
	} -> ((unsigned_char arity, Int32.to_int code), bits);
      | { _ } -> parse_error "invalid lambda entry"

  let parse_lambdas bits =
    bitmatch bits with
      | { count : 32 : nativeendian
	; bits  : -1 : bitstring
	} ->
	  let count = Int32.to_int count in
	  let arr   = Array.make count (0,0) in
	  let rbits = ref bits in
	    for i = 0 to count - 1 do
	      let (entry, bits) = parse_lambda_entry !rbits in
		arr.(i) <- entry;
		rbits := bits
	    done;
	    (arr, parse_eos !rbits)

  let parse_code bits =
    bitmatch bits with
      | { len  : 32 : nativeendian
	; code : (Int32.to_int len * 8) : bitstring
	; bits : -1 : bitstring
	} -> (code, parse_eos bits)

  let parse_footer bits =
    let (sign, bits) = parse_string_entry bits in
      bitmatch bits with
	| { (0xDA15|0xCE03) : 16 : nativeendian
	  ; (0xDA15|0xCE03) : 16 : nativeendian 
	  ; bits : -1 : bitstring, check (bitstring_length bits = 0)
	  } -> sign
	| { _ } -> parse_error "malformed footer"

  let parse bits =
    let (glo, tmp, res, bits) = parse_header bits in
    let (str, bits) = parse_strings bits in
    let (sym, bits) = parse_strings bits in
    let (lam, bits) = parse_lambdas bits in
    let (cod, bits) = parse_code    bits in
    let sign = parse_footer bits in
      (glo, tmp, res, str, sym, lam, cod, sign)

  let from_file f =
    let bits = Bitstring.bitstring_of_file f in
      parse bits

end;;
