(* #load "unix.cma";; *)                (* needed for the interpreter *)
open Printf;;				(* printf *)
open Char;;				(* chr, code *)

exception INVALID_OPCODE of int;;

(* Time wrapper *)
let gettime () =
  Unix.gettimeofday ()
;;

(* Simple read file method *)
let read_file file =
  let buf = Buffer.create 4048 in
  let ic  = open_in file in
  let rec loop () =
    Buffer.add_char buf (input_char ic);
    loop ()
  in
    try
      loop ()
    with End_of_file ->
      close_in ic;
      Buffer.contents buf
;;

(* Performs 't' runs with the interpreter.
 * Before starting the runs a full GC is preformed.
 * This should be enough so that it does not start while
 * interpreting.
 *)
let do_times t f a =
  Gc.full_major ();
  let sum = ref 0 in
  let start = gettime () in
    for i = 1 to t do
      sum := !sum + (f a)
    done;
    let time = gettime () -. start in
      (f a, sum, time)
;;

(* Simple interpreter that reads directly from the string of
 * chars. 'i' is used as an index in the bytecode string and 'r' is
 * the result that is passed on each invocation.
 * 'loop' is in tail position so this uses constant stack space.
 * It is similar to a switch-based interpreter in C as we jump to the
 * same location and start scanning for a bytecode match.
 *)

let interp_str s =
  let rec loop i r =
    match s.[i] with
	'0' -> r
      | '1' -> loop (i+1) (r+1)
      | '2' -> loop (i+1) (r-1)
      | '3' -> loop (i+1) (r lsl 1)
      | '4' -> loop (i+1) ((r lsl 16) lor (r lsr 16))
      | c -> raise (INVALID_OPCODE (code c - code '0'))
  in loop 0 0
;;


(* This interpreter is more or less the same as the above, only now we
 * use a list over a data type of op-codes (opsadt).
 * Again 'loop' is in tail-position and now we pass the tail of
 * op-codes (xs) and the result (r) on each invocation.
 *)

type opsadt
  = Op0
  | Op1
  | Op2
  | Op3
  | Op4
;;

let to_adt ops =
  let rec visit i xs = match i with
      0 -> xs
    | i -> visit (i-1) ((match ops.[i-1] with
			     '0' -> Op0
			   | '1' -> Op1
			   | '2' -> Op2
			   | '3' -> Op3
			   | '4' -> Op4
			   | c -> raise (INVALID_OPCODE (code c - code '0'))
			) :: xs)
  in visit (String.length ops) []
;;

let interp_adt xs =
  let rec loop xs r =
    match xs with
      | [] -> raise (INVALID_OPCODE 999)
      | (x::xs) ->
	  match x with
	      Op0 -> r
	    | Op1 -> loop xs (r+1)
	    | Op2 -> loop xs (r-1)
	    | Op3 -> loop xs (r lsl 1)
	    | Op4 -> loop xs ((r lsl 16) lor (r lsr 16))
  in loop xs 0
;;

(* Indirect Threading.
 * Here all calls go through an opcode table (labels) so we have a
 * single indirection when dispatching to the next opcode function.
 * We give two variants, one on chars and one on integers just to see
 * what difference that might give in OCaml.
 *)

(* Profiling shows that we need to in-line next. To do this we can
 * piggy back on the c-preprocessor and create a traditional
 * C-macro (done with the ``-pp cpp'' flag).
 *
 * Originals:
 *   let next_c () = pc := !pc + 1; labels.(code ops.(!pc)) ()
 *   let next_i () = pc := !pc + 1; labels.(ops.(!pc)) ()
 *   let next_d () = pc := !pc + 1; ops.(!pc) () 
 *)
#define NEXTc pc := !pc + 1; labels.(code ops.(!pc)) ()
#define NEXTi pc := !pc + 1; labels.(ops.(!pc)) ()
#define NEXTd pc := !pc + 1; ops.(!pc) ()

(* Converts a string to a char array *)
let to_chr_arr str =
  let arr = Array.make (String.length str) (chr 0) in
    for i = 0 to String.length str - 1 do
      arr.(i) <- chr (code str.[i] - code '0')
    done;
    arr
;;
(* Converts a string to an int array *)
let to_int_arr str =
  let arr = Array.make (String.length str) 0 in
    for i = 0 to String.length str - 1 do
      arr.(i) <- code str.[i] - code '0'
    done;
    arr
;;

(* 'r' is the result register, 'pc' is our current index into the
 * opcode array, 'next' is a dispatch helper that increments 'pc' and
 * invokes the next opcode handler through the table. 'labels' is an
 * array of opcode handlers that are simple functions of type:
 * unit -> int
 * Notice that only the first label function has an int result, all
 * others have 'next' in tail-position.
 *)
(* #  define next_c pc := !pc + 1; labels.(code ops.(!pc)) *)
let interp_ind_c ops =
  let r  = ref 0 in
  let pc = ref 0 in
  let rec labels =
    [| (fun () -> !r);
       (fun () -> r := !r+1; NEXTc);
       (fun () -> r := !r-1; NEXTc);
       (fun () -> r := !r lsl 1; NEXTc);
       (fun () -> r := (!r lsl 16) lor (!r lsr 16); NEXTc) |]
  in labels.(code ops.(0)) ()
;;

let interp_ind_i ops =
  let r  = ref 0 in
  let pc = ref 0 in
  let rec labels =
    [| (fun () -> !r);
       (fun () -> r := !r+1; NEXTi);
       (fun () -> r := !r-1; NEXTi);
       (fun () -> r := !r lsl 1; NEXTi);
       (fun () -> r := (!r lsl 16) lor (!r lsr 16); NEXTi) |]
  in labels.(ops.(0)) ()
;;

(* Direct Threading.
 * Each opcode is an actual function.
 *
 * 'r' and 'pc' are mutable cells and must be at the top level so
 * interp_dir can reset them.
 *)
let r = ref 0;;
let pc = ref 0;;

(* Convert a bytecode string to an array of opcode labels (functions)
 * Here we locally define as much as possible and the labels are
 * created by lambda closures.
 *)
let to_fun_inn str =
  let emp = (fun () -> 0) in
  let ops = Array.make (String.length str) emp in
  let labels =
    [| (fun () -> !r);
       (fun () -> r := !r+1; NEXTd);
       (fun () -> r := !r-1; NEXTd);
       (fun () -> r := !r lsl 1; NEXTd);
       (fun () -> r := (!r lsl 16) lor (!r lsr 16); NEXTd)
    |]
  in
    for i = 0 to String.length str - 1 do
      ops.(i) <- labels.(code str.[i] - code '0');
    done;
    ops
;;


(* Alternatively we can create everything at the top level.
 * The handler for each opcode is a top level function and the labels
 * array refers to them.
 * Haskell is known to prefer top level declaration when it comes to
 * optimization, so this might help in OCaml?
 *)
let maxl = 4801;;		  (* str length is hard-coded here! *)
let emp = (fun () -> 0);;
let ops = Array.make maxl emp;;
let op0 () = !r ;;
let op1 () = r := !r+1; NEXTd;;
let op2 () = r := !r-1; NEXTd;;
let op3 () = r := !r lsl 1; NEXTd;;
let op4 () = r := (!r lsl 16) lor (!r lsr 16); NEXTd;;
let labels_top = [| op0; op1; op2; op3; op4 |];;
let to_fun_top str =
  for i = 0 to String.length str - 1 do
    ops.(i) <- labels_top.(code str.[i] - code '0');
  done;
  ops
;;

let interp_dir ops =
  (* reset r and pc *)
  r := 0; pc := 0;
  (* invoke the first opcode function (ala. goto) *)
  ops.(0) ()
;;

(* A pure direct threaded version where we pass the result and
 * program counter on each invocation.
 *)

(* Moving ops to the top level avoids an environment lookup in
 * each of the opcode-labels.
 *)
let emp2 = (fun (_:int) (_:int) -> 0);;
let ops2 = Array.make maxl (*String.length str*) emp2;;

let to_fun_pure_named str =
  (* let emp2 = (fun (_:int) (_:int) -> 0) in *)
  (* let ops2 = Array.make maxl (*String.length str*) emp2 in *)

  (* CHANGING THE ORDER OF LABELS REALLY CHANGES PREFORMANCE!!! *)
  let label4 = (fun r pc -> ops2.(pc) ((r lsl 16) lor (r lsr 16)) (pc+1)) in
  let label3 = (fun r pc -> ops2.(pc) (r lsl 1)                   (pc+1)) in
  let label2 = (fun r pc -> ops2.(pc) (r-1)                       (pc+1)) in
  let label1 = (fun r pc -> ops2.(pc) (r+1)                       (pc+1)) in
  let label0 = (fun r pc -> r)                                            in
  (*
    let label0 = (fun r pc -> r)                                            in
    let label1 = (fun r pc -> ops2.(pc) (r+1)                       (pc+1)) in
    let label2 = (fun r pc -> ops2.(pc) (r-1)                       (pc+1)) in
    let label3 = (fun r pc -> ops2.(pc) (r lsl 1)                   (pc+1)) in
    let label4 = (fun r pc -> ops2.(pc) ((r lsl 16) lor (r lsr 16)) (pc+1)) in
  *)
    for i = 0 to String.length str - 1 do
      ops2.(i) <- match (code str.[i] - code '0') with
	  0 -> label0
	| 1 -> label1
	| 2 -> label2
	| 3 -> label3
	| 4 -> label4
	| n -> raise (INVALID_OPCODE n)
    done;
    ops2
;;

let to_fun_pure str =
  (* let emp2 = (fun (_:int) (_:int) -> 0) in *)
  (* let ops2 = Array.make maxl (*String.length str*) emp2 in *)
  let labels =
    [| (fun r pc -> r);
       (fun r pc -> ops2.(pc) (r+1) (pc+1));
       (fun r pc -> ops2.(pc) (r-1) (pc+1));
       (fun r pc -> ops2.(pc) (r lsl 1) (pc+1));
       (fun r pc -> ops2.(pc) ((r lsl 16) lor (r lsr 16)) (pc+1))
    |]
  in
    for i = 0 to String.length str - 1 do
      ops2.(i) <- labels.(code str.[i] - code '0');
    done;
    ops2
;;

(* three arg labels (now including ops) *)
let to_fun_pure3 str =
  (* Changing the order of labels does not do much here... *)

  (* labels are now recursive types: *)
  (* oplabel : \mu X . int -> int -> X array *)
  let oplabel4 = (fun r pc ops -> ops.(pc) ((r lsl 16) lor (r lsr 16)) (pc+1) ops) in
  let oplabel3 = (fun r pc ops -> ops.(pc) (r lsl 1)                   (pc+1) ops) in
  let oplabel2 = (fun r pc ops -> ops.(pc) (r-1)                       (pc+1) ops) in
  let oplabel1 = (fun r pc ops -> ops.(pc) (r+1)                       (pc+1) ops) in
  let oplabel0 = (fun r pc ops -> r)                                               in
  (*
  let oplabel0 = (fun r pc ops -> r)                                               in
  let oplabel1 = (fun r pc ops -> ops.(pc) (r+1)                       (pc+1) ops) in
  let oplabel2 = (fun r pc ops -> ops.(pc) (r-1)                       (pc+1) ops) in
  let oplabel3 = (fun r pc ops -> ops.(pc) (r lsl 1)                   (pc+1) ops) in
  let oplabel4 = (fun r pc ops -> ops.(pc) ((r lsl 16) lor (r lsr 16)) (pc+1) ops) in
  *)
  let ops = Array.make (String.length str) oplabel0 in
    for i = 0 to String.length str - 1 do
      ops.(i) <- match (code str.[i] - code '0') with
	  0 -> oplabel0
	| 1 -> oplabel1
	| 2 -> oplabel2
	| 3 -> oplabel3
	| 4 -> oplabel4
	| n -> raise (INVALID_OPCODE n)
    done;
    ops
;;


let interp_dir_pure ops =
  ops.(0) 0 1
;;

let interp_dir_pure3 ops =
  ops.(0) 0 1 ops
;;

let to_fun_pure_tuple str =
  let emp = (fun ((_:int), (_:int)) -> 0) in
  let ops = Array.make (String.length str) emp in
  let labels =
    [| (fun (r, pc) -> r);
       (fun (r, pc) -> ops.(pc) (r+1, pc+1));
       (fun (r, pc) -> ops.(pc) (r-1, pc+1));
       (fun (r, pc) -> ops.(pc) (r lsl 1, pc+1));
       (fun (r, pc) -> ops.(pc) ((r lsl 16) lor (r lsr 16), pc+1))
    |]
  in
    for i = 0 to String.length str - 1 do
      ops.(i) <- labels.(code str.[i] - code '0');
    done;
    ops
;;
let interp_dir_tuple ops =
  ops.(0) (0, 1)
;;

let to_adt_arr str =
  let arr = Array.make (String.length str) Op0 in
    for i = 0 to (String.length str) - 1 do
      arr.(i) <- (match str.[i] with
		      '0' -> Op0
		    | '1' -> Op1
		    | '2' -> Op2
		    | '3' -> Op3
		    | '4' -> Op4
		    | c -> raise (INVALID_OPCODE (code c - code '0')))
    done;
    arr
;;

let interp_defun ops =
  let rec apply_defun r pc =
    match ops.(pc) with
	Op0 -> r
      | Op1 -> apply_defun (r+1) (pc+1)
      | Op2 -> apply_defun (r-1) (pc+1)
      | Op3 -> apply_defun (r lsl 1) (pc+1)
      | Op4 -> apply_defun ((r lsl 16) lor (r lsr 16)) (pc+1)
  in apply_defun 0 0
;;

exception INVALID_ARGUMENT of string;;
let str2fn ops s = match s with
  | "swstr" -> (fun () -> interp_str ops)
  | "swadt" -> let ops' = to_adt ops in (fun () -> interp_adt ops')
  | "inchr" -> let ops' = to_chr_arr ops in (fun () -> interp_ind_c ops')
  | "inint" -> let ops' = to_int_arr ops in (fun () -> interp_ind_i ops')
  | "drinn" -> let ops' = to_fun_inn ops in (fun () -> interp_dir ops')
  | "drtop" -> let ops' = to_fun_top ops in (fun () -> interp_dir ops')
  | "drpur" -> let ops' = to_fun_pure ops in (fun () -> interp_dir_pure ops')
  | "drpu3" -> let ops' = to_fun_pure3 ops in (fun () -> interp_dir_pure3 ops')
  | "drtup" -> let ops' = to_fun_pure_tuple ops in (fun () -> interp_dir_tuple ops')
  | "drnam" -> let ops' = to_fun_pure_named ops in (fun () -> interp_dir_pure ops')
  | "defun" -> let ops' = to_adt_arr ops in (fun () -> interp_defun ops')
  | _ -> raise (INVALID_ARGUMENT ("Invalid function string: " ^ s))
;;

let main = match Sys.argv with
    [| _; fn; file; times|] ->
      let ops = read_file file in
      let (r,s,t) = do_times (int_of_string times) (str2fn ops fn) () in
	printf "%s,%s,%d,%d,%f" fn file r !s t
  | _ ->
      raise (INVALID_ARGUMENT
	       "Requires 3 arguments: function, file and times.")
;;

(* Notes:
 *
 * Performance seems to be the same for label arrays of type:
 *  (unit -> int) array			(* functions *)
 *  (ref (unit -> int)) array		(* function pointers *)
 * 
 *)
